// Fragment Shader  file "parallaxShader.frag"

#version 330 core
#extension GL_ARB_explicit_uniform_location : enable
//Blinn-Phong Shading
//with three light sources (1. point light, 2-3. spot light)

layout(location = 9) uniform sampler2D colorMap;  //contains texture
layout(location = 11) uniform mat3 reflectionCoefficient;  //1.column ambient 2.column diffuse, 3.column specular
layout(location = 30) uniform float shininess;
layout(location = 22) uniform mat3 spotDirection; //in which direction the spot light shines, in world coordinates
layout(location = 26) uniform float spotCosCutoff[3]; //how big is the light cone
layout(location = 23) uniform float spotExponent[3]; //how strong light decreases from center to walls of light cone
layout(location = 18) uniform sampler2D varianceShadowMap1;
layout(location = 19) uniform sampler2D varianceShadowMap2;
layout(location = 20) uniform sampler2D varianceShadowMap3;
layout(location = 40) uniform int useTexture;
layout(location = 44) uniform sampler2D normalMap;
layout(location = 45) uniform sampler2D heightMap;
layout(location = 46) uniform sampler2D specularMap;
layout(location = 47) uniform int useSpecularMap;

in vec3 normal;			  //normal vector of fragment
in vec3 cameraDirection;  //direction fragment to camera, in tangent space
in vec2 texCoords;
in vec3 color;
in mat3 lightIntensity;   //intensity of light, includes color of light
in mat3 lightDirection;   //direction fragment to light, in tangent space
in mat3 lightDirectionWorld;
in vec4 SM_tex_coords1;	//shadow map coords
in vec4 SM_tex_coords2;	//shadow map coords
in vec4 SM_tex_coords3;	//shadow map coords
//in mat3 spotDirection_tangent;

out vec4 out_color; //final color of fragment


float getLightFraction(sampler2D varianceShadowMap, vec2 texCoord, float dist) {

	vec2 moments = texture(varianceShadowMap,texCoord).rg;
	if (dist <= moments.x)
		return 1.0 ;

	float variance = moments.y - (moments.x*moments.x);
	variance = max(variance,0.00002); //bias?
	
	float d = dist - moments.x;
	float p_max = variance / (variance + d*d);

	return max(p_max,0.2);
}

void main(void) {

	//normalize vectors
	//vec3 normal_norm = normalize(normal);
	vec3 cameraDirection_norm = normalize(cameraDirection);
	vec3 lightDirection_norm;

	//parallax mapping - get height from texture
	float height = texture(heightMap,texCoords.st).r;
	//not sure if height computation is correct...
	float scale = 0.02;
	height = height * scale - 0.5 * scale;
	vec2 newCoords = texCoords.st + (cameraDirection_norm.xy * height);
	//newCoords = texCoords;

	vec3 specularReflectionCoefficient = reflectionCoefficient[2];
	
	if(useSpecularMap == 1){
		float specVal = texture(specularMap, newCoords.st).r;
		specularReflectionCoefficient = vec3(specVal,specVal,specVal);
	}

	//get normal vector from texture
	//parallax mapping
	vec3 normal_norm = texture(normalMap, newCoords.st).rgb * 2.0 - 1.0;
	//normal mapping
	//vec3 normal_norm = texture(normalMap, texCoords.st).rgb * 2.0 - 1.0;
	//no mapping
	//vec3 normal_norm = normalize(normal);

	vec3 SM_tex_coords1_wdivided = (SM_tex_coords1.xyz/SM_tex_coords1.w);
	vec3 SM_tex_coords2_wdivided = (SM_tex_coords2.xyz/SM_tex_coords2.w);
	vec3 SM_tex_coords3_wdivided = (SM_tex_coords3.xyz/SM_tex_coords3.w);

	vec3 diffuse = vec3(0,0,0);
	vec3 specular = vec3(0,0,0);
	vec3 texColor = texture(colorMap,newCoords.st).rgb;
	vec3 surfaceColor = useTexture*texColor+(1-useTexture)*color;
	vec3 ambient = vec3(0.3,0.3,0.3)* reflectionCoefficient[0] * surfaceColor;

	float shadow[3];
	shadow[0] = getLightFraction(varianceShadowMap1, SM_tex_coords1_wdivided.xy,SM_tex_coords1_wdivided.z);
	shadow[1] = getLightFraction(varianceShadowMap2, SM_tex_coords2_wdivided.xy,SM_tex_coords2_wdivided.z);
	shadow[2] = getLightFraction(varianceShadowMap3, SM_tex_coords3_wdivided.xy,SM_tex_coords3_wdivided.z);
	
	if (shadow[0]>0) {

		lightDirection_norm = normalize(lightDirection[0]);

		//diffuse part
		float N_L = dot(normal_norm,lightDirection_norm);
		if (N_L > 0) {

			float spotEffect = dot(normalize(spotDirection[0]), normalize(-lightDirectionWorld[0]));
			//float spotEffect = dot(normalize(spotDirection_tangent[0]), -lightDirection_norm);
			if (spotEffect > spotCosCutoff[0]) {

				spotEffect = max(0,pow(spotEffect, spotExponent[0]));
				diffuse += shadow[0]*reflectionCoefficient[1] * ((surfaceColor*spotEffect*lightIntensity[0]) * N_L);

				//specular part
				vec3 halfway = cameraDirection_norm+lightDirection_norm;
				float N_H = dot(normal_norm,normalize(halfway));
				if (N_H > 0) {
					float highlight = pow(N_H,shininess);
					specular += shadow[0]*specularReflectionCoefficient * (spotEffect*lightIntensity[0]* highlight);
				}
			}
		}
	}
	
	if (shadow[1]>0) {

		lightDirection_norm = normalize(lightDirection[1]);

		//diffuse part
		float N_L = dot(normal_norm,lightDirection_norm);
		if (N_L > 0) {

			float spotEffect = dot(normalize(spotDirection[1]), normalize(-lightDirectionWorld[1]));
			//float spotEffect = dot(normalize(spotDirection_tangent[1]), -lightDirection_norm);
			if (spotEffect > spotCosCutoff[1]) {

				spotEffect = max(0,pow(spotEffect, spotExponent[1]));
				diffuse += shadow[1]*reflectionCoefficient[1] * ((surfaceColor*spotEffect*lightIntensity[1]) * N_L);

				//specular part
				vec3 halfway = cameraDirection_norm+lightDirection_norm;
				float N_H = dot(normal_norm,normalize(halfway));
				if (N_H > 0) {
					float highlight = pow(N_H,shininess);
					specular += shadow[1]*specularReflectionCoefficient * (spotEffect*lightIntensity[1]* highlight);
				}
			}
		}
	}
	
	if (shadow[2]>0) {

		lightDirection_norm = normalize(lightDirection[2]);

		//diffuse part
		float N_L = dot(normal_norm,lightDirection_norm);
		if (N_L > 0) {

			float spotEffect = dot(normalize(spotDirection[2]), normalize(-lightDirectionWorld[2]));
			//float spotEffect = dot(normalize(spotDirection_tangent[2]), -lightDirection_norm);
			if (spotEffect > spotCosCutoff[2]) {
			
				spotEffect = max(0,pow(spotEffect, spotExponent[2]));
				diffuse += shadow[2]*reflectionCoefficient[1] * ((surfaceColor*spotEffect*lightIntensity[2]) * N_L);

				//specular part
				vec3 halfway = cameraDirection_norm+lightDirection_norm;
				float N_H = dot(normal_norm,normalize(halfway));
				if (N_H > 0) {
					float highlight = pow(N_H,shininess);
					specular += shadow[2]*specularReflectionCoefficient * (spotEffect*lightIntensity[2]* highlight);
				}
			}
		}
	}

	//adding all parts together to get the result
	out_color = vec4(ambient + diffuse + specular,1.0);

}
